/*
 * AI Noise Reduction Example 
 * This firmware programs the SPU with an AI Noise Reduction algorithm that filters out background noise while preserving
 * human speech. Pressing the button turns on the algorithm, where the green LED indicates that the algorithm is on. Press the
 * button again to turn off the algorithm and hear the unfiltered audio output. The volume slider on the side of the Tympan controls
 * the output gain of the headphones. Various debugging information is also displayed in the serial terminal.
 *
 * Before loading this firmware, do the following:
 * 1) Insert an SD card with the files 0PROG_A and 0PROG_D corresponding to the example AINR model. 0PROG_P can optionally be included
 *    to change parameters from the default (e.g. for power optimization).
 * 2) Make sure the EVK is connected according to the quick start guide.
 * 3) Attach headphones to the black audio output jack. An external microphone can optionally be plugged into the red audio jack. Note that
 *    you must set USE_EXTERNAL_MIC 1 in 0PROG_P to use the external mic.
 * 4) The Button2 library by Lennart Hennigs, Teensy 1.57.2, and the Tympan library are required to run this firmware. Install them in 
 * the Arduino Board and Library Managers.
 *
 */

#include <Tympan_Library.h>    //Tympan library
#include <SD.h>                //SD card library
#include "ProcessAudio_F32.h"  //Audio processing algorithm (other tab)
#include <AudioStream.h>       //Audio streaming library
#include "spu.h"               //Femtosense SPU library
#include <Button2.h>           //Button library

//Default parameters (will be overwritten by 0PROG_P file if it exists)
//Descriptions of these parameters are in spu.h
uint32_t SPI_SPEED = 4000000;
float VCO_MHZ = 100;
int CORE_GATING_EN = 0;
int CORE0_READY_US = 0;
int CORE_TRANSFER_US = 0;
int SAFETY_MARGIN_US = 0;
int TOTAL_WAIT_TIME_US = 0;  
int USE_MEMORY_FSM = 0;
int CLKOD = 1;
int TUNE_DEBUG = 0;
float INPUT_GAIN_DB = 25.0f;      //gain on the microphone (0-47.5dB in 0.5dB steps)
float OUTPUT_MAX_GAIN_DB = 0.0f;  //max gain of output
int USE_EXTERNAL_MIC = 0;         //use external mic (0=no, 1=yes)


//IO pins for onboard LEDs
#define LED_GREEN 15
#define LED_RED 16
#define BUTTON 25

//Serial BAUD rate
#define SERIAL_BAUD 115200

//configure audio
AudioSettings_F32 audio_settings(SAMPLE_RATE_HZ, HOP_SIZE);

//create audio library objects for handling the audio
Tympan myTympan(TympanRev::E, audio_settings);  //Initialize Tympan
AudioInputI2S_F32 i2s_in(audio_settings);       //Digital audio from Codec.
ProcessAudio_F32 effect1(audio_settings);       //Attach audio processing algorithm in ProcessAudio.h. This runs every hop.
AudioEffectGain_F32 gain1(audio_settings);      //This is for the volume knob
AudioOutputI2S_F32 i2s_out(audio_settings);     //Digital audio to Codec.

//audio connections
AudioConnection_F32 patchCord1(i2s_in, 0, effect1, 0);  //connect the Left input to the processing algorithm
AudioConnection_F32 patchCord2(effect1, 0, gain1, 0);   //connect to the gain effect
AudioConnection_F32 patchCord3(gain1, 0, i2s_out, 0);   //connect to the Left output
AudioConnection_F32 patchCord4(gain1, 0, i2s_out, 1);   //also connect to the Right output (mono -> stereo)

//setup button
Button2 button;

//current volume
float volume_dB = 0.0;  //this value will be overridden by volume knob position

//current loopback state
int loopback_mode = 1;  // 1=loopback mode, 0=AINR mode. Start in loopback mode.

//runs once at startup
void setup() {

  //Initialize serial port
  Serial.begin(SERIAL_BAUD);
  delay(1);
  Serial.println("Setup started.");

  //Program model parameters
  //Mount SD card
  if (!SD.begin(BUILTIN_SDCARD)) {
    error("SD card initialization failed! Check that card is inserted correctly.", 4, 1, 0);
  }
  if (!setModelParameters()) {
    error("Model parameters on SD card are invalid", 4, 1, 0);
  }

  //Check that AUDIO_BLOCK_SAMPLES is set correctly
  //This file is located at: C:\Users\<USERNAME>\AppData\Local\Arduino15\packages\teensy\hardware\avr\1.57.2\cores\teensy4\AudioStream.h or similar
  if (AUDIO_BLOCK_SAMPLES < HOP_SIZE)
    error("You must increase AUDIO_BLOCK_SAMPLES in AudioStream.h to be >= HOP_SIZE", 2, 1, 0);

  //Disable flash chip that shares SPI bus with SPU on EVK board
  pinMode(SS_FLASH, OUTPUT);
  digitalWrite(SS_FLASH, HIGH);

  //Enable SPU interrupt pin
  pinMode(SPU_INT, INPUT);

  //Initialize SPU and run integrity and clock checks
  int init_error = initializeSPU();
  if (init_error == 0)
    error("SPU integrity check failed. Check wiring.", 3, 1, 1);
  if (init_error == -1)
    error("SPU clock check failed. Check clock jumper and signal.", 5, 1, 1);

  //Program SPU from SD Card
  if (!programModel())
    error("Model could not be loaded from SD card.", 4, 1, 1);

  //turn off PLL and clocks until needed
  SPUClockPower(0);
  SPUCorePower(0, 0);

  //Start audio feed
  AudioMemory_F32(20, audio_settings);               //allocate the audio memory
  myTympan.enable();                                 //activate Codec
if (USE_EXTERNAL_MIC) {                         //attach input
    myTympan.inputSelect(TYMPAN_INPUT_JACK_AS_MIC);  
    Serial.println("Connected to External Mic!");
  } else {
    myTympan.inputSelect(TYMPAN_INPUT_ON_BOARD_MIC);
  }
  myTympan.setInputGain_dB(INPUT_GAIN_DB);           //set input gain
  myTympan.volume_dB(0);                             //set output gain. This will be set later by the volume knob

  //set initial volume and mic detect states
  servicePotentiometer(millis(), 0);

  //setup button
  button.begin(BUTTON);
  button.setTapHandler(tap);

  //Flash LED green to indicate device is ready
  flashLED(LED_GREEN, 3);
  Serial.println("Setup complete.");
}

//runs after setup and repeats forever.
void loop() {

  //handle volume change
  servicePotentiometer(millis(), 100);  //service the volume knob every 100ms

  //check and see if AINR loop exited because of an error
  if (loopback_mode != effect1.getLoopbackMode()) {

    //go to loopback mode and permanently error.
    loopback_mode = effect1.getLoopbackMode();
    updateMode();
    error("SPU processing error, restart required. Check power optimization parameters and PLL frequency. If this error occured on first frame, check that the model file is correct.", 6, 1, 1);
  }

  //handle button press
  button.loop(); //if button is not working, try increasing SPU clock speed so that interrupt asserts sooner.
}

//Reads files 0PROG_A and 0PROG_D from root directory of SD card and programs model into SPU
//Returns 1 if successful, else returns 0
int programModel() {

  //Setup files
  File addressFile;
  File dataFile;
  addressFile = SD.open("0PROG_A");
  dataFile = SD.open("0PROG_D");

  //make sure files exist
  if (!addressFile) {
    Serial.println("Cannot find file 0PROG_A on SD card");
    return 0;
  }
  if (!dataFile) {
    Serial.println("Cannot find file 0PROG_D on SD card");
    return 0;
  }

  //Read file in chunks of 512 lines (this is faster than reading line by line)
  int i = 0;

  unsigned long addresses[512];
  while (1) {
    char address[9];
    addressFile.readBytes(address, 8);
    address[8] = '\0';
    char CR_LF = addressFile.read();  //consume CR or LF
    if (CR_LF == '\015')              //if CR, there will also be LF
      addressFile.read();             //consume LF

    addresses[i % 512] = strtoul(address, '\0', 16);

    if ((i % 512 == 511) || (!addressFile.available())) {

      //print every 512 lines to show progress
      Serial.print("Programming from file: Line ");
      Serial.print(i);
      Serial.println("...");

      //for each line in file
      for (int j = 0; j <= (i % 512); j++) {
        char data[9];
        dataFile.readBytes(data, 8);
        data[8] = '\0';
        CR_LF = dataFile.read();
        if (CR_LF == '\015')
          dataFile.read();
        unsigned long datas;
        datas = strtoul(data, '\0', 16);

        //If using FSM, put memories in FSM mode instead of ON mode
        if (USE_MEMORY_FSM)
          if (((uint32_t)addresses[j] <= 0x98) && ((uint32_t)addresses[j] >= 0x34) && ((uint32_t)datas == 0x11098))
            datas = 0x15098;

        //write to SPU
        SPU_Write(0x01, (uint32_t)addresses[j], 1, (uint32_t*)&datas);
      }

      //file has ended
      if (!addressFile.available())
        break;
    }
    i++;
  }

  //completed successfully
  return 1;
}

//Flases LED flashCount number of times.
void flashLED(int LEDPin, int flashCount) {
  for (int i = 0; i < flashCount; i++) {
    digitalWrite(LEDPin, HIGH);
    delay(150);
    digitalWrite(LEDPin, LOW);
    delay(150);
  }
}

//Prints error message and ends code execution while flashing red LED flashCount number of times.
void error(String errorMessage, int flashCount, int stopProgram, int dumpRegisters) {

  //print error message
  Serial.print("ERROR");
  Serial.print(flashCount);
  Serial.print(": ");
  Serial.println(errorMessage);

  //dump registers for debugging
  if (dumpRegisters) {
    Serial.println("SPI Register Dump:");
    uint32_t SPUBuffer[8];
    SPU_Read(0x00, 0, 8, SPUBuffer);
    for (int i = 0; i < 8; i++) {
      Serial.print("SPI Addr ");
      Serial.print(i);
      Serial.print(": 0x");
      Serial.println(SPUBuffer[i], HEX);
    }
  }

  //loop forever
  while (1) {
    flashLED(LED_RED, flashCount);
    delay(1000);

    //loop once
    if (!stopProgram)
      break;
  }
}

//Checks position of volume knob and updates output gain accordingly. Updates are limited to once per updatePeriod_millis ms
void servicePotentiometer(unsigned long curTime_millis, unsigned long updatePeriod_millis) {

  //keep track of previous values
  static unsigned long lastUpdate_millis = 0;
  static float prev_val = 0;

  //has enough time passed to update everything?
  if (curTime_millis < lastUpdate_millis) lastUpdate_millis = 0;     //handle wrap-around of the clock
  if ((curTime_millis - lastUpdate_millis) > updatePeriod_millis) {  //is it time to update?

    //read potentiometer knob
    float val = float(myTympan.readPotentiometer()) / 1023.0;  //0.0 to 1.0
    val = (1.0 / 9.0) * (float)((int)(9.0 * val + 0.5));       //quantize so that it doesn't chatter...0 to 1.0

    //update volume if value has changed
    if (abs(val - prev_val) > 0.05) {  //is it different than before?
      prev_val = val;                  //save the value for comparison for the next time around

      //choose the desired gain value based on the knob setting
      const float min_gain_dB = -20.0, max_gain_dB = 30.0;          //set desired gain range
      volume_dB = min_gain_dB + (max_gain_dB - min_gain_dB) * val;  //computed desired gain value in dB

      //clip max gain to OUTPUT_MAX_GAIN_DB set in 0PROG_P file
      if (volume_dB > OUTPUT_MAX_GAIN_DB) 
        volume_dB = OUTPUT_MAX_GAIN_DB;

      //update volume via gain setting
      gain1.setGain_dB(volume_dB);  //set the gain
      Serial.print("Updating volume: Digital Gain dB = ");
      Serial.println(volume_dB);  //print text to Serial port
    }
    lastUpdate_millis = curTime_millis;
  }
}


//Called whenever the button is pressed
void tap(Button2& button) {

  //toggle between loopback mode and AINR mode
  loopback_mode = (loopback_mode) ? 0 : 1;

  updateMode();
}

//Toggles between loopback mode and AINR mode. Call after loopback_mode is changed.
void updateMode() {

  //enter loopback mode
  if (loopback_mode) {

    //stop AINR
    effect1.setLoopbackMode(loopback_mode);

    //turn off clock and cores
    SPUClockPower(0);
    SPUCorePower(0, 0);

    Serial.println("Switched to Loopback Mode...");

    //turn off green LED
    digitalWrite(LED_GREEN, !loopback_mode);

    //enter AINR
  } else {
    //turn on clock
    if (!SPUClockPower(1)) {
      error("SPU PLL unable to lock. Check clock jumper and signal.", 5, 1, 1);
    }

    Serial.println("Switched to AINR Mode...");

    //turn on green LED
    digitalWrite(LED_GREEN, !loopback_mode);

    //start AINR
    effect1.setLoopbackMode(loopback_mode);
  }
}

//Reads file 0PROG_P from root directory of SD card and sets model parameters
//If the file does not exist, a non-blocking error is shown, and the default values are not changed.
//Returns 1 if successful and 0 if the parameters read are invalid.
int setModelParameters() {

  //load file
  File parametersFile;
  parametersFile = SD.open("0PROG_P");
  if (!parametersFile) {
    error("0PROG_P file not found, using default parameters...", 1, 0, 0);
    return 1;
  }

  //Keep track of core gating parameters. All must be set if CORE_GATING_EN == 1.
  int setCORE_GATING_EN = -1;
  int setCORE0_READY_US = -1;
  int setCORE_TRANSFER_US = -1;
  int setSAFETY_MARGIN_US = -1;

  while (parametersFile.available()) {

    //read space delimited file
    String parameterNameStr = parametersFile.readStringUntil(' ');
    String parameterValueStr = parametersFile.readStringUntil('\n');

    //set parameters and print parsed values
    String strval;
    Serial.print("0PROG_P: ");
    if (parameterNameStr == "SPI_SPEED") {
      SPI_SPEED = parameterValueStr.toInt();
      strval = String(SPI_SPEED);
    } else if (parameterNameStr == "VCO_MHZ") {
      VCO_MHZ = parameterValueStr.toFloat();
      strval = String(VCO_MHZ);
    } else if (parameterNameStr == "CORE_GATING_EN") {
      setCORE_GATING_EN = parameterValueStr.toInt();
      strval = String(setCORE_GATING_EN);
    } else if (parameterNameStr == "CORE0_READY_US") {
      setCORE0_READY_US = parameterValueStr.toInt();
      strval = String(setCORE0_READY_US);
    } else if (parameterNameStr == "CORE_TRANSFER_US") {
      setCORE_TRANSFER_US = parameterValueStr.toInt();
      strval = String(setCORE_TRANSFER_US);
    } else if (parameterNameStr == "SAFETY_MARGIN_US") {
      setSAFETY_MARGIN_US = parameterValueStr.toInt();
      strval = String(setSAFETY_MARGIN_US);
    } else if (parameterNameStr == "TOTAL_WAIT_TIME_US") {
      TOTAL_WAIT_TIME_US = parameterValueStr.toInt();
      strval = String(TOTAL_WAIT_TIME_US);
    } else if (parameterNameStr == "USE_MEMORY_FSM") {
      USE_MEMORY_FSM = parameterValueStr.toInt();
      strval = String(USE_MEMORY_FSM);
    } else if (parameterNameStr == "INPUT_GAIN_DB") {
      INPUT_GAIN_DB = parameterValueStr.toFloat();
      strval = String(INPUT_GAIN_DB);
    } else if (parameterNameStr == "CLKOD") {
      CLKOD = parameterValueStr.toInt();
      strval = String(CLKOD);
    } else if (parameterNameStr == "OUTPUT_MAX_GAIN_DB") {
      OUTPUT_MAX_GAIN_DB = parameterValueStr.toFloat();
      strval = String(OUTPUT_MAX_GAIN_DB);
    } else if (parameterNameStr == "TUNE_DEBUG") {
      TUNE_DEBUG = parameterValueStr.toInt();
      strval = String(TUNE_DEBUG);
    } else if (parameterNameStr == "USE_EXTERNAL_MIC") {
      USE_EXTERNAL_MIC = parameterValueStr.toInt();
      strval = String(USE_EXTERNAL_MIC);
    } else {
      strval = "<This is an unknown parameter!>";
    }
    Serial.print(parameterNameStr);
    Serial.print(" = ");
    Serial.println(strval);
  }

  //Check CLKOD
  if ((CLKOD < 1 || CLKOD > 16) || (CLKOD != 1 && CLKOD % 2 == 1)) {
    Serial.println("Power tuning parameters error-  CLKOD must be: 1 <= CLKOD <= 16 and either 1 or even.");
    return 0;
  }

  //check core gating parameters
  if ((setCORE_GATING_EN == 1)) {
    //check that all parameters needed are included
    if ((setCORE0_READY_US > -1) && (setCORE_TRANSFER_US > -1) && (setSAFETY_MARGIN_US > -1)) {
      //set core gating parameters
      CORE_GATING_EN = setCORE_GATING_EN;
      CORE0_READY_US = setCORE0_READY_US;
      CORE_TRANSFER_US = setCORE_TRANSFER_US;
      SAFETY_MARGIN_US = setSAFETY_MARGIN_US;

      //parameters invalid
      if (!((!CORE_GATING_EN) || (((CORE0_READY_US - SAFETY_MARGIN_US) + (CORE_TRANSFER_US + 2 * SAFETY_MARGIN_US)) <= TOTAL_WAIT_TIME_US))) {
        Serial.println("Power tuning parameters error- individual delays are greater than total wait time allowed");
        return 0;
      }

      Serial.println("Core gating enabled!");

      //parmeters invalid
    } else {
      Serial.println("Core gating is enabled, but not all required parameters were set.");
      return 0;
    }
  }

  //return 1 for success.
  return 1;
}