/* This audio streaming and processing architecture is based on the Tympan example sketch "MyAudioAlgorithm"
 */

#include "AudioStream_F32.h"
#include "spu.h"

//Class definition
class ProcessAudio_F32 : public AudioStream_F32 {
public:

  //constructor
  ProcessAudio_F32(const AudioSettings_F32 &settings)
    : AudioStream_F32(1, inputQueueArray_f32) {
    ;
  };

  //This is called every time a new chunk of audio is received
  void update(void) {

    //get audio block
    audio_block_f32_t *audio_block;
    audio_block = AudioStream_F32::receiveWritable_f32();
    if (!audio_block) return;

    //process audio
    process(audio_block);

    ///transmit the block and release memory
    AudioStream_F32::transmit(audio_block);
    AudioStream_F32::release(audio_block);
  }

  //This is called to process the actual block of audio samples
  void process(audio_block_f32_t *audio_block) {

    //only process audio with AINR algorithm if not in loopback mode
    if (!loopback_mode) {

      //create buffer for communicating with SPU
      const int SPU_BUFFER_SIZE = HOP_SIZE / 4 * 2 + 2;
      uint32_t SPUBuffer[SPU_BUFFER_SIZE];

      //set route and PC data
      SPUBuffer[0] = 0x00000000;  //route
      SPUBuffer[1] = 0x00000000;  //PC (anything works in this case)


      //pack samples into buffer to send to SPU
      int j = 0;
      for (int i = 0; i < (HOP_SIZE / 4 * 4); i += 4) {
        int32_t samples[4];
        for (int k = 0; k < 4; k++)
          samples[k] = (audio_block->data[i + k]) * 32767;  //convert to integer

        SPUBuffer[2 + j++] = (samples[0] & 0x0000FFFF) | ((samples[1] << 16) & 0xFFFF0000);
        SPUBuffer[2 + j++] = (samples[2] & 0x0000FFFF) | ((samples[3] << 16) & 0xFFFF0000);
      }

      if (CORE_GATING_EN) {
        SPUCorePower(1, 0);  //turn on core0, leave core1 off for now
      } else {
        SPUCorePower(1, 1);  //turn both cores on
      }

      //send first N-4 samples with AXIS
      SPU_Write(0x02, 0, SPU_BUFFER_SIZE - 2, SPUBuffer);

      //turn off clock divider in preparation for data processing
      SPUSetOD(1);

      //send final 4 samples with AXIS_LAST. This starts the processing.
      SPU_Write(0x03, 0, 2, &SPUBuffer[SPU_BUFFER_SIZE - 2]);

      if (CORE_GATING_EN) {
        delayMicroseconds(CORE0_READY_US - SAFETY_MARGIN_US);  //wait for core0 to finish processing
      }

      //turn on core1 if using core gating
      SPUCorePower(1, 1);

      if (CORE_GATING_EN) {
        delayMicroseconds(CORE_TRANSFER_US + SAFETY_MARGIN_US * 2);  //wait for core0-to-core1 transfer to finish
      }


      if (CORE_GATING_EN) {
        SPUCorePower(0, 1);  //turn off core0, leave core1 on
      } else {
        SPUCorePower(1, 1);
      }

      //use delay to save IO power polling intrerupt
      if (!TUNE_DEBUG) {
        if (CORE_GATING_EN) {
          delayMicroseconds(TOTAL_WAIT_TIME_US - (CORE0_READY_US - SAFETY_MARGIN_US) - (CORE_TRANSFER_US + SAFETY_MARGIN_US * 2));
        } else {
          delayMicroseconds(TOTAL_WAIT_TIME_US);
        }
      }

      //wait for interrupt to pull data back from SPU
      unsigned long timer = micros();
      int triedtorecover_coreson = 0;  //if using core gating, this tracks whether recovery was tried if no interrupt was seen
      while (digitalRead(SPU_INT) == LOW) {
        SPI.transfer(0x00);  //poll interrupt by pulsing SPI SCK (no SS is asserted)
        if (CORE_GATING_EN) {
          if (((micros() - timer) > 500) && (!triedtorecover_coreson)) {  //if no interrupt is seen after 500us, turn on both cores and see if processing recovers
            Serial.print(millis());
            Serial.println(": Core not finished? Attempting to recover...");
            SPUCorePower(1, 1);
            triedtorecover_coreson = 1;
          }
        }

        if ((micros() - timer) > 5000) {  //If no interrupt is seen after 5000us, there was a serious problem. exit.
          Serial.print(millis());
          Serial.println(": Could not Recover, exiting AINR.");
          setLoopbackMode(1);

          return;
        }
      }

      if (triedtorecover_coreson) {  //turning on both cores successfully recovered from issue!
        Serial.print(millis());
        Serial.println(": Recovered!");
      }

      //for debuging: Print how long it took to see interrupt
      if (TUNE_DEBUG) {
        Serial.println(micros() - timer);
      }

      //make sure processing is done. This may not be needed.
      delayMicroseconds(10);

      //processing is done now, turn on clock divider to save power
      SPUSetOD(CLKOD);

      //read response from SPU
      SPU_Read(0x03, 0, SPU_BUFFER_SIZE, SPUBuffer);  //the response from this model is the AINR filtered audio

      //You can turn off core1 early here, but it only saves 36us
      //SPUCorePower(0, 0);

      //unpack samples from buffer received from SPU, and place them in i2s output buffer
      j = 0;
      for (int i = 2; i < SPU_BUFFER_SIZE; i += 2) {
        int16_t samples[4];
        samples[0] = (SPUBuffer[i] & 0x0000FFFF) >> 0;       //unpack 16 bit sample
        audio_block->data[j++] = 0.0000305185 * samples[0];  //convert to floating point

        samples[1] = ((SPUBuffer[i] & 0xFFFF0000) >> 16);
        audio_block->data[j++] = 0.0000305185 * samples[1];

        samples[2] = (SPUBuffer[i + 1] & 0x0000FFFF) >> 0;
        audio_block->data[j++] = 0.0000305185 * samples[2];

        samples[3] = ((SPUBuffer[i + 1] & 0xFFFF0000) >> 16);
        audio_block->data[j++] = 0.0000305185 * samples[3];
      }
    }
  }

  //sets loopback_mode tracker
  void setLoopbackMode(int setting) {
    loopback_mode = setting;
  }

  //gets loopback_mode tracker
  int getLoopbackMode() {
    return loopback_mode;
  }

private:
  audio_block_f32_t *inputQueueArray_f32[1];  //memory pointer for the input to this module

  int loopback_mode = 1;  //keep track of loopback_mode
};
