#include "spu.h"

/*Initializes the SPU
 *returns 0 if registers cannot be read
 *returns -1 if PLL does not lock
 *returns 1 if everything is okay
 */
int initializeSPU() {

  //initialize ports of SPI data transfer pins
  SPI.begin();

  //setup reset and SPI SS pins
  pinMode(SPU_RESET_PIN, OUTPUT);
  pinMode(SS, OUTPUT);

  //reset SPU
  SPUReset(SPU_RESET_PIN);

  //Check that SPU registers can be read, return 0 if not.
  if (!SPUIntegrityCheck())
    return 0;

  //setup SPU PLL, return -1 if PLL lock failure
  if (!SPUPLLSetup())
    return -1;

  //reset SPU again - the hard reset only works properly after a power cycle or while PLL is locked.
  SPUReset(SPU_RESET_PIN);

  //setup SPU PLL again
  if (!SPUPLLSetup())
    return -1;

  //default state has clock divided
  SPUSetOD(CLKOD);

  return 1;
}

//Resets the SPU using the reset pin
void SPUReset(int resetPin) {
  Serial.print("Resetting SPU");

  //hold reset for 1us
  digitalWrite(resetPin, HIGH);
  delayMicroseconds(1);
  digitalWrite(resetPin, LOW);
  delayMicroseconds(1);
  Serial.println(".");
}

//Sets up the SPU PLL registers
int SPUPLLSetup() {

  //buffer to write to SPU
  uint32_t SPUBuffer[4];

  float VCO = VCO_MHZ;
  uint32_t CLKF = VCO * 1000000 / 32768 - 1;  //for 32.768KHz crystal
  uint32_t CLKODexact = 0x00;  //CLKOD = 1
  uint32_t CLKR = 0x00;
  uint32_t BWADJ = (CLKF + 1) / 2 - 1;

  //these fields are offset by 1 bit
  SPUBuffer[0] = BWADJ << 1;
  SPUBuffer[1] = CLKODexact << 1;
  SPUBuffer[2] = CLKF << 1;
  SPUBuffer[3] = CLKR << 1;
  SPU_Write(0x00, 12, 4, SPUBuffer);

  //turn on PLL and check for lock
  return SPUClockPower(1);
}

/* Request to read SPU data.
 * instruction: SPU instruction 
 * address: start address of data requested
 * length: length of data requested
 * data: buffer for returned data from SPU
 */
void SPU_Read(byte instruction, uint32_t address, int length, uint32_t* data) {
  int MessageLength = 1 + 8 + 4;    //bytes required for request
  byte WriteBuffer[MessageLength];  //holds instruction, start/end addresses, and 4 stall bytes

  WriteBuffer[0] = (instruction << 1);  //set wr_en to 0 for reading

  //set start and end addresses
  for (int i = 0; i < 4; i++) {
    WriteBuffer[1 + i] = (byte)((address << (24 - i * 8)) >> 24);
    WriteBuffer[5 + i] = (byte)(((address + length * 4) << (24 - i * 8)) >> 24);
  }
  WriteBuffer[MessageLength - 4] = 0x00;  //stall bytes
  WriteBuffer[MessageLength - 3] = 0x00;  //stall bytes
  WriteBuffer[MessageLength - 2] = 0x00;  //stall bytes
  WriteBuffer[MessageLength - 1] = 0x00;  //stall bytes

  //SPI transaction to request read
  SPI.beginTransaction(SPISettings(SPI_SPEED, MSBFIRST, SPI_MODE0));
  digitalWrite(SS, LOW);                     //assert SPU SS
  SPI.transfer(WriteBuffer, MessageLength);  //no response expected for write, so no SPI receive data buffer is set
  digitalWrite(SS, HIGH);                    //deassert SPU SS
  SPI.endTransaction();

  //setup read buffer to receive data
  MessageLength = (4 * length);
  byte ReadBuffer[MessageLength];

  //SPI transaction to read data
  SPI.beginTransaction(SPISettings(SPI_SPEED, MSBFIRST, SPI_MODE0));
  digitalWrite(SS, LOW);                    //assert SPU SS
  SPI.transfer(ReadBuffer, MessageLength);  //response will be written to ReadBuffer. Anything can be written here, so no SPI send data buffer is set
  digitalWrite(SS, HIGH);                   //deassert SPU SS
  SPI.endTransaction();

  //store data that was read into function output buffer
  for (int i = 0; i < length; i++) {
    data[i] = ReadBuffer[i * 4 + 0] + ((uint32_t)ReadBuffer[i * 4 + 1] << 8) + ((uint32_t)ReadBuffer[i * 4 + 2] << 16) + ((uint32_t)ReadBuffer[i * 4 + 3] << 24);
  }
}

/* Write data to SPU.
 * instruction: SPU instruction 
 * address: start address of write location
 * length: length of data to write
 * data: buffer of data to write to SPU
 */
void SPU_Write(byte instruction, uint32_t address, int length, uint32_t* data) {

  int MessageLength = (1 + 8 + 4 * length + 2);  //bytes required for request
  byte WriteBuffer[MessageLength];               //holds instruction, start/end addresses, data, and 2 stall bytes

  WriteBuffer[0] = (instruction << 1) | 0x01;  //set wr_en to 1 for writing

  //set start and end addresses
  for (int i = 0; i < 4; i++) {
    WriteBuffer[1 + i] = (byte)((address << (24 - i * 8)) >> 24);
    WriteBuffer[5 + i] = (byte)(((address + length * 4) << (24 - i * 8)) >> 24);
  }

  //load data into write buffer
  for (int i = 0; i < length; i++) {
    WriteBuffer[9 + i * 4 + 0] = (byte)((data[i] << 24) >> 24);
    WriteBuffer[9 + i * 4 + 1] = (byte)((data[i] << 16) >> 24);
    WriteBuffer[9 + i * 4 + 2] = (byte)((data[i] << 8) >> 24);
    WriteBuffer[9 + i * 4 + 3] = (byte)(data[i] >> 24);
  }
  WriteBuffer[MessageLength - 2] = 0x00;  //stall bytes
  WriteBuffer[MessageLength - 1] = 0x00;  //stall bytes

  //SPI transaction to write data
  SPI.beginTransaction(SPISettings(SPI_SPEED, MSBFIRST, SPI_MODE0));
  digitalWrite(SS, LOW);                     //assert SPU SS
  SPI.transfer(WriteBuffer, MessageLength);  //no response expected for write, so no SPI receive data buffer is set
  digitalWrite(SS, HIGH);                    //deassert SPU SS
  SPI.endTransaction();
}

/* Tests basic SPU register reads. This must be run after reset, before any SPI registers are changed.
 *returns 0 if default registers cannot be read
 *returns 1 if default registers can be read
 */
int SPUIntegrityCheck() {

  //simply check that we can read the default values of the config registers
  uint32_t SPUBuffer[7];
  uint32_t SPUDefaults[7] = { 0x00, 0x248, 0x00, 0x3E6, 0x02, 0x7CE, 0x00 };

  //read values from SPU
  SPU_Read(0x00, 0, 7, SPUBuffer);

  //compare to defaults
  int OKAY = 1;
  for (int i = 0; i < 7; i++)
    if (SPUBuffer[i] != SPUDefaults[i])
      OKAY = 0;

  return OKAY;
}

/* Turn on and off the PLL and oscillator
 * on: 1- turns on PLL, 0- turns off PLL 
 * For on==1, returns 1 if PLL is locked and 0 if not
 * For on==0, returns 1
 */
int SPUClockPower(int on) {
  uint32_t SPUBuffer[2];

  //turn on pll
  if (on) {
    SPUBuffer[0] = 0x448;  //turn on oscillator
    SPUBuffer[1] = 0x0;    //turn on PLL
    SPU_Write(0x00, 4, 2, SPUBuffer);

    //reset PLL to update
    SPUBuffer[0] = 1;
    SPU_Write(0x00, 8, 1, SPUBuffer);
    SPUBuffer[0] = 0;
    SPU_Write(0x00, 8, 1, SPUBuffer);

    //wait up to 1000ms for PLL Lock
    uint32_t locked = 0;
    unsigned long timer = millis();
    while ((!locked) && ((millis() - timer < 1000))) {
      SPU_Read(0x00, 28, 1, SPUBuffer);
      locked = SPUBuffer[0] & 0x00000001;
    }
    return locked;

    //turn off pll
  } else {
    SPUBuffer[0] = 0x048;  //turn off oscillator
    SPUBuffer[1] = 0x20;   //turn off PLL
    SPU_Write(0x00, 4, 2, SPUBuffer);

    //reset PLL to update
    SPUBuffer[0] = 0x21;
    SPU_Write(0x00, 8, 1, SPUBuffer);
    SPUBuffer[0] = 0x20;
    SPU_Write(0x00, 8, 1, SPUBuffer);

    return 1;
  }

  //return for invalid input
  return -1;
}

/* Toggles SPU core power
 * core0: 1- turns on core, 0- turns off core 
 * core1: 1- turns on core, 0- turns off core 
 */
void SPUCorePower(int core0, int core1) {
  uint32_t SPUBuffer[1];

  //set disable_cores registers
  SPUBuffer[0] = (!(core1 > 0) << 2) | (!(core0 > 0) << 1);
  SPU_Write(0x00, 0, 1, SPUBuffer);
}

/* Sets SPU PLL output divider. Warning: Do not change clock divider during processing!
 * divider: clock divider 
 */
void SPUSetOD(byte divider) {
  uint32_t SPUBuffer[1];

  //set output divider register
  SPUBuffer[0] = ((divider - 1) & 0x0F) << 1;
  SPU_Write(0x00, 16, 1, SPUBuffer);
}