#ifndef SPU_H
#define SPU_H

#include "Arduino.h"
#include <SPI.h>

//Audio Parameters
#define SAMPLE_RATE_HZ 16000.0f  //Other frequencies available listed in the table in AudioOutputI2S_F32

//Hop size (number of samples) of SPU input
#define HOP_SIZE 64 //I2S buffer size to fill before processing. For 16ms AINR model, use 128. For 8ms AINR model, use 64

//SPI speed
extern uint32_t SPI_SPEED;

//Power tuning parameters:
extern float VCO_MHZ;             //VCO crequency (MHz)  

extern int CORE_GATING_EN;        //enable core gating (0:off 1:on)
extern int CORE0_READY_US;        //Core 0 processing time
extern int CORE_TRANSFER_US;      //Core-to-core transfer time. 
extern int SAFETY_MARGIN_US;      //Safety margin around core-to-core transfer to account for inconsistant execution time
extern int TOTAL_WAIT_TIME_US;    //total time available to wait for processing.

extern int USE_MEMORY_FSM;        //use FSM memory mode (0:off 1:on)

extern int CLKOD;                 //This output divider is used during IO transfers when the cores are not processing data.

extern int TUNE_DEBUG;            //TUNE_DEBUG==1 prints SPU processing timer.

//IO Pins
#define SPU_RESET_PIN 32  //SPU reset signal
#define SPU_INT 6         //SPU interrupt
#define SS_FLASH 24       //flash chip SS on the SPU EVK board


/*Initializes the SPU
 *returns 0 if registers cannot be read
 *returns -1 if PLL does not lock
 *returns 1 if everything is okay
 */
int initializeSPU();

//Resets the SPU using the reset pin
void SPUReset(int resetPin);

//Sets up the SPU PLL registers
int SPUPLLSetup();

/* Request to read SPU data.
 * instruction: SPU instruction 
 * address: start address of data requested
 * length: length of data requested
 * data: buffer for returned data from SPU
 */
void SPU_Read(byte instruction, uint32_t address, int length, uint32_t* data);

/* Write data to SPU.
 * instruction: SPU instruction 
 * address: start address of write location
 * length: length of data to write
 * data: buffer of data to write to SPU
 */
void SPU_Write(byte instruction, uint32_t address, int length, uint32_t* data);

/* Tests basic SPU register reads. This must be run after reset, before any SPI registers are changed.
 *returns 0 if default registers cannot be read
 *returns 1 if default registers can be read
 */
int SPUIntegrityCheck();

/* Turn on and off the PLL and oscillator
 * on: 1- turns on PLL, 0- turns off PLL 
 * For on==1, returns 1 if PLL is locked and 0 if not
 * For on==0, returns 1
 */
int SPUClockPower(int on);

/* Toggles SPU core power
 * core0: 1- turns on core, 0- turns off core 
 * core1: 1- turns on core, 0- turns off core 
 */
void SPUCorePower(int core0, int core1);

/* Sets SPU PLL output divider. Warning: Do not change clock divider during processing!
 * divider: clock divider 
 */
void SPUSetOD(byte divider);

#endif