/* Wakeword Detection Example 
 * This firmware programs the SPU with a keyword spotting algorithm that listens for the phrase "Hey Snips." When the phrase
 * is heard, the red and green LEDs flash, and "Hey Snips!" is displayed on the serial terminal along with the confidence metric.
 * Various debugging information is also displayed in the serial terminal.
 *
 * Before running this firmware, do the following:
 * 1) Insert an SD card with the files 0PROG_A and 0PROG_D corresponding to the example WWD model. 0PROG_P can optionally be included
 *    to change parameters from the default (e.g. for power optimization).
 * 2) Wire the SPU to the Tympan according to the setup tutorial. Note: The button is not needed for this.
 * 3) Ensure that AUDIO_BLOCK_SAMPLES in AudioStream.h is set to at least the value of HOP_SIZE in spu.h below. This file is likely located at:
 * C:\Users\<USERNAME>\AppData\Local\Arduino15\packages\teensy\hardware\avr\1.57.2\cores\teensy4\AudioStream.h or similar
 *
 */

#include <Tympan_Library.h>    //Tympan library
#include <SD.h>                //SD card library
#include "ProcessAudio_F32.h"  //Audio processing algorithm
#include <AudioStream.h>       //Audio streaming library
#include "spu.h"               //Femtosense SPU library

//Default parameters (will be overwritten by 0PROG_P file if it exists)
//More descriptions of these parameters are in spu.h
uint32_t SPI_SPEED = 4000000;
float VCO_MHZ = 100;
unsigned int TOTAL_WAIT_TIME_US = 0;
int USE_MEMORY_FSM = 0;
int CLKOD = 1;
int TUNE_DEBUG = 0;
float INPUT_GAIN_DB = 47.5f;  //gain on the microphone (0-47.5dB in 0.5dB steps)
int USE_EXTERNAL_MIC = 0;     //use external mic (0=no, 1=yes)
int DISABLE_CORE1 = 0;        //disable core1 if not needed  (0=core1 enabled, 1=core1 disabled)

//IO pins for onboard LEDs
#define LED_GREEN 15
#define LED_RED 16

//IO pin for the flash chip SS on the SPU EVK board.
#define SS_FLASH 24

//Serial BAUD rate
#define SERIAL_BAUD 115200

//configure audio
//Note: There is a bug in Arduino audio libraries that prevents 8Khz sampling from working. Use 16KHz instead and then decimate.
AudioSettings_F32 audio_settings(SAMPLE_RATE_HZ * 2, HOP_SIZE * 2);  

//create audio library objects for handling the audio
Tympan myTympan(TympanRev::E, audio_settings);  //Initialize Tympan
AudioInputI2S_F32 i2s_in(audio_settings);       //Digital audio from Codec.
ProcessAudio_F32 effect1(audio_settings);       //Attach audio processing algorithm in ProcessAudio.h. This runs every hop.

//connect the Left input to the processing algorithm
AudioConnection_F32 patchCord1(i2s_in, 0, effect1, 0);

//runs once at startup
void setup() {

  //Initialize serial port
  Serial.begin(SERIAL_BAUD);
  delay(1);
  Serial.println("Setup started.");

  //Program model parameters
  //Mount SD card
  if (!SD.begin(BUILTIN_SDCARD)) {
    error("SD card initialization failed! Check that card is inserted correctly.", 4, 1, 0);
  }
  if (!setModelParameters()) {
    error("Model parameters on SD card are invalid", 4, 1, 0);
  }

  //Check that AUDIO_BLOCK_SAMPLES is set correctly
  if (AUDIO_BLOCK_SAMPLES < (HOP_SIZE * 2))
    error("You must increase AUDIO_BLOCK_SAMPLES in AudioStream.h to be >= HOP_SIZE", 2, 1, 0);

  //Disable flash chip that shares SPI bus with SPU on EVK board
  pinMode(SS_FLASH, OUTPUT);
  digitalWrite(SS_FLASH, HIGH);

  //Enable SPU interrupt pin
  pinMode(SPU_INT, INPUT);

  //Initialize SPU and run integrity and clock checks
  int init_error = initializeSPU();
  if (init_error == 0)
    error("SPU integrity check failed. Check wiring.", 3, 1, 1);
  if (init_error == -1)
    error("SPU clock check failed. Check clock jumper and signal.", 5, 1, 1);

  //Program SPU from SD Card
  if (!programModel())
    error("Model could not be loaded from SD card.", 4, 1, 1);

  //turn off core1. Most WWD models only use core0
  if (DISABLE_CORE1) {
    SPUCorePower(1, 0);
    Serial.println("Disabling core1!");
  }

  //Turn on clock divider
  SPUSetOD(CLKOD);

  //Start audio feed
  AudioMemory_F32(20, audio_settings);  //allocate the audio memory
  myTympan.enable();                    //activate Codec
  if (USE_EXTERNAL_MIC) {               //attach input
    myTympan.inputSelect(TYMPAN_INPUT_JACK_AS_MIC);  
    Serial.println("Connected to External Mic!");
  } else {
    myTympan.inputSelect(TYMPAN_INPUT_ON_BOARD_MIC); 
  }
  myTympan.setInputGain_dB(INPUT_GAIN_DB);  //set input gain

  //Flash LED green to indicate device is ready
  flashLED(LED_GREEN, 3);
  Serial.println("Setup complete.");
}

//Reads files 0PROG_A and 0PROG_D from root directory of SD card and programs SPU
//Returns 1 if successful, else returns 0
int programModel() {

  //Setup files
  File addressFile;
  File dataFile;
  addressFile = SD.open("0PROG_A");
  dataFile = SD.open("0PROG_D");

  //make sure files exist
  if (!addressFile) {
    Serial.println("Cannot find file 0PROG_A on SD card");
    return 0;
  }
  if (!dataFile) {
    Serial.println("Cannot find file 0PROG_D on SD card");
    return 0;
  }

  //Read file in chunks of 512 lines (this is faster than reading line by line)
  int i = 0;

  unsigned long addresses[512];
  while (1) {
    char address[9];
    addressFile.readBytes(address, 8);
    address[8] = '\0';
    char CR_LF = addressFile.read();  //consume CR or LF
    if (CR_LF == '\015')              //if CR, there will also be LF
      addressFile.read();             //consume LF

    addresses[i % 512] = strtoul(address, '\0', 16);

    if ((i % 512 == 511) || (!addressFile.available())) {

      //print every 512 lines to show progress
      Serial.print("Programming from file: Line ");
      Serial.print(i);
      Serial.println("...");

      //for each line in file
      for (int j = 0; j <= (i % 512); j++) {
        char data[9];
        dataFile.readBytes(data, 8);
        data[8] = '\0';
        CR_LF = dataFile.read();
        if (CR_LF == '\015')
          dataFile.read();
        unsigned long datas;
        datas = strtoul(data, '\0', 16);

        //If using FSM, put memories in FSM mode instead of ON mode
        if (USE_MEMORY_FSM)
          if (((uint32_t)addresses[j] <= 0x98) && ((uint32_t)addresses[j] >= 0x34) && ((uint32_t)datas == 0x11098))
            datas = 0x15098;

        //write to SPU
        SPU_Write(0x01, (uint32_t)addresses[j], 1, (uint32_t*)&datas);
      }

      //file has ended
      if (!addressFile.available())
        break;
    }
    i++;
  }

  //completed successfully
  return 1;
}

//Flases LED flashCount number of times.
void flashLED(int LEDPin, int flashCount) {
  for (int i = 0; i < flashCount; i++) {
    digitalWrite(LEDPin, HIGH);
    delay(150);
    digitalWrite(LEDPin, LOW);
    delay(150);
  }
}

//Prints error message and ends code execution while flashing red LED flashCount number of times.
void error(String errorMessage, int flashCount, int stopProgram, int dumpRegisters) {

  //print error message
  Serial.print("ERROR");
  Serial.print(flashCount);
  Serial.print(": ");
  Serial.println(errorMessage);

  //dump registers for debugging
  if (dumpRegisters) {
    Serial.println("SPI Register Dump:");
    uint32_t SPUBuffer[8];
    SPU_Read(0x00, 0, 8, SPUBuffer);
    for (int i = 0; i < 8; i++) {
      Serial.print("SPI Addr ");
      Serial.print(i);
      Serial.print(": 0x");
      Serial.println(SPUBuffer[i], HEX);
    }
  }

  //loop forever
  while (1) {
    flashLED(LED_RED, flashCount);
    delay(1000);

    //loop once
    if (!stopProgram)
      break;
  }
}

//runs after setup and repeats forever.
void loop() {
  //check and see if ProcessAudio_F32.h exited because of an error
  if (!effect1.getIsResponding()) {
    error("SPU processing error, restart required. Check power optimization parameters and PLL frequency. If this error occured on first frame, check that the model file is correct.", 6, 1, 1);
  }

  //Serial.println("loop()");  //you can print something here to see if audio processing is complete before the next frame comes in (nothing will print if not)
}

//Reads file 0PROG_P from root directory of SD card and sets model parameters
//If the file does not exist, a non-blocking error is shown, and the default values are not changed.
//Returns 1 if successful and 0 if the parameters read are invalid.
int setModelParameters() {

  //load file
  File parametersFile;
  parametersFile = SD.open("0PROG_P");
  if (!parametersFile) {
    error("0PROG_P file not found, using default parameters...", 1, 0, 0);
    return 1;
  }

  while (parametersFile.available()) {

    //read space delimited file
    String parameterNameStr = parametersFile.readStringUntil(' ');
    String parameterValueStr = parametersFile.readStringUntil('\n');

    //set parameters and print parsed values
    String strval;
    Serial.print("0PROG_P: ");
    if (parameterNameStr == "SPI_SPEED") {
      SPI_SPEED = parameterValueStr.toInt();
      strval = String(SPI_SPEED);
    } else if (parameterNameStr == "VCO_MHZ") {
      VCO_MHZ = parameterValueStr.toFloat();
      strval = String(VCO_MHZ);
    } else if (parameterNameStr == "TOTAL_WAIT_TIME_US") {
      TOTAL_WAIT_TIME_US = parameterValueStr.toInt();
      strval = String(TOTAL_WAIT_TIME_US);
    } else if (parameterNameStr == "USE_MEMORY_FSM") {
      USE_MEMORY_FSM = parameterValueStr.toInt();
      strval = String(USE_MEMORY_FSM);
    } else if (parameterNameStr == "INPUT_GAIN_DB") {
      INPUT_GAIN_DB = parameterValueStr.toFloat();
      strval = String(INPUT_GAIN_DB);
    } else if (parameterNameStr == "CLKOD") {
      CLKOD = parameterValueStr.toInt();
      strval = String(CLKOD);
    } else if (parameterNameStr == "USE_EXTERNAL_MIC") {
      USE_EXTERNAL_MIC = parameterValueStr.toInt();
      strval = String(USE_EXTERNAL_MIC);
    } else if (parameterNameStr == "TUNE_DEBUG") {
      TUNE_DEBUG = parameterValueStr.toInt();
      strval = String(TUNE_DEBUG);
    } else if (parameterNameStr == "DISABLE_CORE1") {
      DISABLE_CORE1 = parameterValueStr.toInt();
      strval = String(DISABLE_CORE1);
    } else {
      strval = "<This is an unknown parameter!>";
    }
    Serial.print(parameterNameStr);
    Serial.print(" = ");
    Serial.println(strval);
  }

  //Check CLKOD
  if ((CLKOD < 1 || CLKOD > 16) || (CLKOD != 1 && CLKOD % 2 == 1)) {
    Serial.println("Power tuning parameters error-  CLKOD must be: 1 <= CLKOD <= 16 and either 1 or even.");
    return 0;
  }

  //return 1 for success.
  return 1;
}
